/*
 * Copyright (c) 2001-2003, David Janssens
 * Copyright (c) 2002-2003, Yannick Verschueren
 * Copyright (c) 2003-2005, Francois Devaux and Antonin Descampe
 * Copyright (c) 2005, Herv Drolon, FreeImage Team
 * Copyright (c) 2002-2005, Communications and remote sensing Laboratory, Universite catholique de Louvain, Belgium
 * Copyright (c) 2007 Stefan Westerfeld
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "t1xx.h"
#include "opj_includes.h"
#include <assert.h>

/* code duplication */
static int t1xx_getctxno_mag(opj_t1_t *t1, int f) {
	return t1->lut_ctxno_mag[(f & T1_SIG_OTH) | (((f & T1_REFINE) != 0) << 11)];
}

static int t1xx_getctxno_zc(opj_t1_t *t1, int f, int orient) {
	return t1->lut_ctxno_zc[(orient << 8) | (f & T1_SIG_OTH)];
}

static int t1xx_getctxno_sc(opj_t1_t *t1, int f) {
	return t1->lut_ctxno_sc[(f & (T1_SIG_PRIM | T1_SGN)) >> 4];
}

static int t1xx_getspb(opj_t1_t *t1, int f) {
	return t1->lut_spb[(f & (T1_SIG_PRIM | T1_SGN)) >> 4];
}

enum T1XXPassType
{
  T1XX_PASS_REF,
  T1XX_PASS_SIG,
  T1XX_PASS_CLN
};
/* --------------------------------- magnitude refinement pass -----------------------------------*/

template<char type, int bpno>
static void t1xx_dec_refpass_step(opj_t1_t *t1, int *fp, int *dp)
{
	const int flag = *fp;

	if ((flag & (T1_SIG | T1_VISIT)) == T1_SIG) {
		opj_mqc_t *mqc = t1->mqc;	/* MQC component */
		int v;

		if (type == T1_TYPE_RAW) {
			opj_raw_t *raw = t1->raw;	/* RAW component */

			mqc_setcurctx(mqc, t1xx_getctxno_mag(t1, flag));	/* ESSAI */
			v = raw_decode(raw);
		} else {
			v = mqc_setcurctx_and_decode(mqc, t1xx_getctxno_mag(t1, flag));
		}

		const int one = 1 << bpno;
		const int poshalf = one >> 1;
		const int neghalf = bpno > 0 ? -poshalf : -1;
		const int pn[4] = { neghalf, poshalf, poshalf, neghalf };

		*dp += pn[((*dp < 0) << 1) + v];
		*fp |= T1_REFINE;
	}
}				/* VSC and  BYPASS by Antonin  */

template<char type, int bpno>
void t1xx_dec_refpass (opj_t1_t *t1, int w, int h)
{
	for (int k = 0; k < h; k += 4) {
		if (k + 3 < h)
		{
			for (int i = 0; i < w; i++)
			{
				t1xx_dec_refpass_step<type, bpno> (t1, &t1->flags[k + 1][1 + i], &t1->data[k    ][i]);
				t1xx_dec_refpass_step<type, bpno> (t1, &t1->flags[k + 2][1 + i], &t1->data[k + 1][i]);
				t1xx_dec_refpass_step<type, bpno> (t1, &t1->flags[k + 3][1 + i], &t1->data[k + 2][i]);
				t1xx_dec_refpass_step<type, bpno> (t1, &t1->flags[k + 4][1 + i], &t1->data[k + 3][i]);
			}
		}
		else /* unlikely */
		{
			for (int i = 0; i < w; i++) {
				for (int j = k; j < k + 4 && j < h; j++) {
					t1xx_dec_refpass_step<type, bpno> (t1, &t1->flags[1 + j][1 + i], &t1->data[j][i]);
				}
			}
		}
	}
}

/* --------------------------------- significance pass -----------------------------------*/

static void t1xx_updateflags(int *fp, int s, int w) {
	int *np = fp - w;
	int *sp = fp + w;
#if 0
	np[-1] |= T1_SIG_SE;
	np[1] |= T1_SIG_SW;
	sp[-1] |= T1_SIG_NE;
	sp[1] |= T1_SIG_NW;
	*np |= T1_SIG_S;
	*sp |= T1_SIG_N;
	fp[-1] |= T1_SIG_E;
	fp[1] |= T1_SIG_W;
	if (s) {
		*np |= T1_SGN_S;
		*sp |= T1_SGN_N;
		fp[-1] |= T1_SGN_E;
		fp[1] |= T1_SGN_W;
	}
#endif
	static const int new_flags[] =
	{
	  T1_SIG_S, 
	  T1_SIG_S | T1_SGN_S,
	  T1_SIG_E,
	  T1_SIG_E | T1_SGN_E,
	  T1_SIG_W,
	  T1_SIG_W | T1_SGN_W,
	  T1_SIG_N,
	  T1_SIG_N | T1_SGN_N
	};

	np[-1] |= T1_SIG_SE;
    np[0]  |= new_flags[s];
	np[1]  |= T1_SIG_SW;
    
	fp[-1] |= new_flags[s + 2];
    fp[1]  |= new_flags[s + 4];

	sp[-1] |= T1_SIG_NE;
    sp[0]  |= new_flags[s + 6];
	sp[1]  |= T1_SIG_NW;
}

template<char type, int bpno>
static void t1xx_dec_sigpass_step(opj_t1_t *t1, int *fp, int *dp, int orient)
{
	const int flag = (*fp);
	if ((flag & T1_SIG_OTH) && !(flag & (T1_SIG | T1_VISIT))) {
		const int one = 1 << bpno;
		const int half = one >> 1;
		const int oneplushalf = one | half;
		const int oh[2] = { oneplushalf, -oneplushalf };
			
		if (type == T1_TYPE_RAW) {
		    opj_raw_t *raw = t1->raw;	/* RAW component */

			if (raw_decode(raw)) {
				const int v = raw_decode(raw);	/* ESSAI */

				*dp = oh[v];
				t1xx_updateflags(fp, v, t1->flags_width);
				*fp |= T1_SIG | T1_VISIT;
				return;
			}
		} else {
			opj_mqc_t *mqc = t1->mqc;	/* MQC component */

			if (mqc_setcurctx_and_decode (mqc, t1xx_getctxno_zc(t1, flag, orient))) {
				const int v = mqc_setcurctx_and_decode (mqc, t1xx_getctxno_sc(t1, flag)) ^ t1xx_getspb(t1, flag);

				*dp = oh[v];
				t1xx_updateflags(fp, v, t1->flags_width);
				*fp |= T1_SIG | T1_VISIT;
				return;
			}
		}
		*fp |= T1_VISIT;
	}
}

template<char type, int bpno>
void t1xx_dec_sigpass(opj_t1_t *t1, int w, int h, int orient) {
	for (int k = 0; k < h; k += 4) {
		if (k + 3 < h)
		{
			for (int i = 0; i < w; i++)
			{
				t1xx_dec_sigpass_step<type, bpno> (t1, &t1->flags[k + 1][1 + i], &t1->data[k][i], orient);
				t1xx_dec_sigpass_step<type, bpno> (t1, &t1->flags[k + 2][1 + i], &t1->data[k + 1][i], orient);
				t1xx_dec_sigpass_step<type, bpno> (t1, &t1->flags[k + 3][1 + i], &t1->data[k + 2][i], orient);
				t1xx_dec_sigpass_step<type, bpno> (t1, &t1->flags[k + 4][1 + i], &t1->data[k + 3][i], orient);
			}
		}
		else /* unlikely */
		{
			for (int i = 0; i < w; i++) {
				for (int j = k; j < k + 4 && j < h; j++) {
					t1xx_dec_sigpass_step<type, bpno> (t1, &t1->flags[j + 1][1 + i], &t1->data[j][i], orient);
				}
			}
		}
	}
}

template<char type, int bpno>
void t1_dec_clnpass_step(opj_t1_t *t1, int *fp, int *dp, int orient, int partial) {
	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
	
	const int flag = *fp;
	if (partial) {
		goto LABEL_PARTIAL;
	}
	if (!(flag & (T1_SIG | T1_VISIT))) {
		if (mqc_setcurctx_and_decode(mqc, t1xx_getctxno_zc(t1, flag, orient))) {
LABEL_PARTIAL:
			const int v = mqc_setcurctx_and_decode(mqc, t1xx_getctxno_sc(t1, flag)) ^ t1xx_getspb(t1, flag);
			const int one = 1 << bpno;
			const int half = one >> 1;
			const int oneplushalf = one | half;
			const int oh[2] = { oneplushalf, -oneplushalf };

			*dp = oh[v];
			t1xx_updateflags(fp, v, t1->flags_width);
			*fp |= T1_SIG;
		}
	}
	*fp &= ~T1_VISIT;
}

template<char type, int bpno>
void t1xx_dec_clnpass(opj_t1_t *t1, int w, int h, int orient)
{
	int agg, runlen;

	opj_mqc_t *mqc = t1->mqc;	/* MQC component */
	
	for (int k = 0; k < h; k += 4) {
		if (k + 3 < h) {
			for (int i = 0; i < w; i++) {
				agg = !(t1->flags[1 + k][1 + i] & (T1_SIG | T1_VISIT | T1_SIG_OTH)
					|| t1->flags[1 + k + 1][1 + i] & (T1_SIG | T1_VISIT | T1_SIG_OTH)
					|| t1->flags[1 + k + 2][1 + i] & (T1_SIG | T1_VISIT | T1_SIG_OTH)
					|| t1->flags[1 + k + 3][1 + i] & (T1_SIG | T1_VISIT | T1_SIG_OTH));
				
				if (agg)
				{
					if (!mqc_setcurctx_and_decode(mqc, T1_CTXNO_AGG)) {
						continue;
					}
					runlen = mqc_setcurctx_and_decode(mqc, T1_CTXNO_UNI);
					runlen = (runlen << 1) | mqc_decode(mqc);
				} else {
					runlen = 0;
				}
				for (int j = k + runlen; j < k + 4; j++) {
					t1_dec_clnpass_step<type, bpno> (t1, &t1->flags[1 + j][1 + i], &t1->data[j][i], orient, agg && (j == k + runlen));
				}
			}
		} else { /* unlikely */
			for (int i = 0; i < w; i++) {
				for (int j = k; j < k + 4 && j < h; j++) {
					t1_dec_clnpass_step<type, bpno> (t1, &t1->flags[1 + j][1 + i], &t1->data[j][i], orient, 0);
				}
			}
		}
	}
}


template<char type, int bpno>
void t1xx_dec_pass (opj_t1_t *t1, int w, int h, int orient, T1XXPassType pass)
{
	switch (pass)
	{
		case T1XX_PASS_SIG: t1xx_dec_sigpass<type, bpno> (t1, w, h, orient);
							return;
		case T1XX_PASS_REF: t1xx_dec_refpass<type, bpno> (t1, w, h);
							return;
		case T1XX_PASS_CLN: t1xx_dec_clnpass<type, bpno> (t1, w, h, orient);
							return;
	}
	printf ("bad pass: %d\n", pass);
}

template<char type>
void t1xx_dec_pass (opj_t1_t *t1, int w, int h, int orient, int bpno, T1XXPassType pass)
{
	switch (bpno)
	{
		case 0: t1xx_dec_pass<T1_TYPE_MQ, 0> (t1, w, h, orient, pass);
				break;
		case 1: t1xx_dec_pass<T1_TYPE_MQ, 1> (t1, w, h, orient, pass);
				break;
		case 2: t1xx_dec_pass<T1_TYPE_MQ, 2> (t1, w, h, orient, pass);
				break;
		case 3: t1xx_dec_pass<T1_TYPE_MQ, 3> (t1, w, h, orient, pass);
				break;
		case 4: t1xx_dec_pass<T1_TYPE_MQ, 4> (t1, w, h, orient, pass);
				break;
		case 5: t1xx_dec_pass<T1_TYPE_MQ, 5> (t1, w, h, orient, pass);
				break;
		case 6: t1xx_dec_pass<T1_TYPE_MQ, 6> (t1, w, h, orient, pass);
				break;
		case 7: t1xx_dec_pass<T1_TYPE_MQ, 7> (t1, w, h, orient, pass);
				break;
		case 8: t1xx_dec_pass<T1_TYPE_MQ, 8> (t1, w, h, orient, pass);
				break;
		case 9: t1xx_dec_pass<T1_TYPE_MQ, 9> (t1, w, h, orient, pass);
				break;
		case 10: t1xx_dec_pass<T1_TYPE_MQ, 10> (t1, w, h, orient, pass);
				 break;
		case 11: t1xx_dec_pass<T1_TYPE_MQ, 11> (t1, w, h, orient, pass);
				 break;
		case 12: t1xx_dec_pass<T1_TYPE_MQ, 12> (t1, w, h, orient, pass);
				 break;
		default: printf ("bad bpno %d\n", bpno);
	}
}

void t1xx_dec_pass (opj_t1_t *t1, int w, int h, int orient, int bpno, char type, T1XXPassType pass)
{
	switch (type)
	{
		case T1_TYPE_RAW:	t1xx_dec_pass<T1_TYPE_RAW> (t1, w, h, orient, bpno, pass);
							return;
		case T1_TYPE_MQ:	t1xx_dec_pass<T1_TYPE_MQ>  (t1, w, h, orient, bpno, pass);
							return;
		default:			printf ("bad type %d\n", type);
	}
}

void t1xx_dec_refpass (opj_t1_t *t1, int w, int h, int bpno, char type)
{
	const int orient = 0;
	t1xx_dec_pass (t1, w, h, orient, bpno, type, T1XX_PASS_REF);
}

void t1xx_dec_sigpass(opj_t1_t *t1, int w, int h, int bpno, int orient, char type)
{
	t1xx_dec_pass (t1, w, h, orient, bpno, type, T1XX_PASS_SIG);
}

void t1xx_dec_clnpass(opj_t1_t *t1, int w, int h, int bpno, int orient)
{
	t1xx_dec_pass (t1, w, h, orient, bpno, T1_TYPE_MQ, T1XX_PASS_CLN);
}
